/* MoarNES source - NES.c

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to simulate the internal NES architecture,
   such as memory mapping, redirecting read/writes to I/O registers, etc.
*/

#include <stdio.h>
#include "moarnes.h"

/*
   Nintendo Entertainment System: CPU memory map

   $0000-$07FF = 2 KB internal RAM
   $0800-$1FFF = Mirrors of $0000-$07FF
   $2000-$2007 = PPU registers
   $2008-$3FFF = Mirrors of $2000-$2007
   $4000-$4017 = APU and I/O registers
   $4018-$FFFF = Cartridge PRG ROM, cartridge PRG RAM, and mapper registers

   $FFFA = NMI vector
   $FFFC = Reset vector
   $FFFE = IRQ/BRK vector

   In some cartridges, $6000-$7FFF is SRAM.

*/

uint32_t totalframes = 0;
uint8_t NESRAM[0x800]; //2 KB internal RAM
uint32_t cpugoal = 0, cpuadjust;
uint16_t sprtablesave;
uint8_t persistenttext[64];
uint16_t texttimer = 0;
struct cheat_s cheat[3];

extern uint8_t gamegenie;

extern struct cart_s cartridge; //struct is declared in rom.c
extern struct PPU_s *PPU;
extern struct OAM_s *OAM;

extern struct map4_s *map4;

extern uint8_t outputNES[240][256];

extern uint8_t savestatefilename[MAX_PATH];

#ifndef _WIN32
	struct timeval timestr;
#else
	LARGE_INTEGER dummytimer;
#endif
uint64_t timerfreq, starttimer, endtimer, difftimer, curtimer, nextframe, latestframe, frametime;
double doubletimer;
extern uint64_t nextsamplecycle, sampleticks, nextseqcycle, seqtickgap;
extern uint8_t dorecord;
extern SDL_Surface *screen;

bool running = false;
bool paused = false;
bool doloadstate = false;
bool dosavestate = false;

uint8_t read6502(uint16_t addr) {
	uint8_t retval, cnum;

	if (addr < 0x2000) retval = NESRAM[addr & 0x7FF];
	else if (addr < 0x4000) retval = readPPUregs(0x2000 | (addr & 7));
	else if (addr < 0x4018) {
		switch (addr) {
			case 0x4016: retval = joyread(0); break;
			case 0x4017: retval = joyread(1); break;
			default: retval = readAPU(addr); break;
		}
	}
	else if (addr < 0x8000) {
		if (cartridge.hasSRAM == true) {
			retval = cartridge.SRAM[addr & 0x1FFF];
		} else retval = 0; //open data bus if no cartidge SRAM
	} else {
		//if we get to this point, it's a read from the cartridge
		addr &= 0x7FFF;
		retval = cartridge.PRGbank[addr>>10][addr&1023];
	}
	if (gamegenie == 2) {
		for (cnum=0; cnum<3; cnum++) {
			if ((cheat[cnum].valid) && (addr == cheat[cnum].addr)) if (cheat[cnum].docompare == 1) {
				if (retval == cheat[cnum].compare) retval = cheat[cnum].replace;
			} else {
				retval = cheat[cnum].replace;
			}
		}
	}

	return(retval);
}

void write6502(uint16_t addr, uint8_t value) {
	uint16_t tempsprite;
	if ((cartridge.mapper>0) && (addr>0x7FFF)) if (mapperwrite(addr, value)) return;
	if ((gamegenie == 1) && (addr>0x7FFF)) { geniewrite(addr, value); return; }
	if (addr < 0x2000) { NESRAM[addr & 0x7FF] = value; return; }
	if (addr < 0x4000) { writePPUregs(0x2000 | (addr & 7), value); return; }
	if (addr < 0x4018) {
		if (addr == 0x4014) {
			for (tempsprite=0; tempsprite<256; tempsprite++) {
				OAM->RAM[(OAM->addr+tempsprite)&255] = read6502(value*256+tempsprite);
			}
			return;
		}
		if (addr == 0x4016) joystrobe();
			else writeAPU(addr, value);
		return;
	}

	if (addr < 0x8000) {
		if (cartridge.hasSRAM == true) {
			cartridge.SRAM[addr & 0x1FFF] = value;
		} else return; //open data bus if no cartidge SRAM
	}

	//if we get to this point, it's an attempted write to the cartridge.
	//obviously, being ROM, nothing would actually get written anywhere.
}

void execframe(bool skiprender) {
	uint16_t scanline;

	if (cartridge.mapper == 4) map4->irqcounter = map4->irqlatch;
	PPU->vblank = 0;
	PPU->sprzero = 0;
	PPU->sprover = 0;
	sprtablesave = PPU->sprtable;
	if (PPU->bgvisible) {
		exec6502(101);
		PPU->addr = PPU->tempaddr;
		exec6502(13);
	} else exec6502(114);
	for (scanline=0; scanline<240; scanline++) {
		if (cartridge.mapper == 4) {
			if (map4->irqenable && map4->irqcounter == 0) irq6502();
			if (map4->irqcounter == 0) map4->irqcounter = map4->irqlatch;
			if (PPU->bgvisible || PPU->sprvisible) {
				map4irqdecrement();
				//if (map4->irqcounter > 0) map4->irqcounter--;
				//	else map4->irqcounter = map4->irqlatch;
			}
		}

#ifdef PIXEL_ACCURACY
		if ((scanline%3)==0) exec6502(8);
#else
		if ((scanline%3)==0) exec6502(86);
			else exec6502(85);
#endif
			renderscanline(scanline, skiprender);

		if (PPU->bgvisible) PPU->addr = (PPU->addr & 0xFBE0) | (PPU->tempaddr & 0x041F);

		exec6502(28);
	}

	exec6502(340);

	PPU->vblank = 1;
	if (PPU->nmivblank) nmi6502();

	exec6502(1930);

	totalframes++;
}

extern uint8_t VRAM[16384];
extern uint16_t changescreenmode, screenmode;
extern bool blitting;
uint32_t framedrops = 0;
void execNES() {
	uint8_t skipblit;
	reset6502();
	memset(&NESRAM[0], 0, 2048);

	getfreq(timerfreq);
	//if (dorecord) initrecord();

	frametime = (uint64_t)((double)timerfreq / (double)60.0);
	nextsamplecycle = getclockticks() + sampleticks;
	nextseqcycle = getclockticks() + seqtickgap;
	getclock(starttimer);
    nextframe = starttimer + frametime;
	while (running) {
		skipblit = 0;
		for (;;) {
			getclock(curtimer);
					if (((int64_t)curtimer - (int64_t)nextframe) > ((int64_t)frametime)) {
						skipblit = 1;
						framedrops++;
						nextframe = curtimer - frametime;
						//nextframe += frametime;
						//if (nextframe > curtimer) nextframe = curtimer;
//						printf("Frames dropped: %lu\n", framedrops);
						break;
					}
			if (curtimer >= nextframe) break;
			SDL_Delay(1);
		}
		/*if (!skipblit) execframe(false);
			else execframe(true);*/
		execframe(false);
		if (texttimer) {
			drawtext(&persistenttext[0], 0, 232);
			texttimer--;
		}
		//if ((curtimer - nextframe) > (frametime << 6)) nextframe = curtimer + frametime;
		nextframe += frametime;
		processinput();
		if (!skipblit && !changescreenmode) blitframe(&outputNES[0][0]);
		if (changescreenmode && !blitting) {
			setscreenmode(changescreenmode);
			changescreenmode = 0;
			if (screenmode == 5) {
				sprintf(&persistenttext[0], "Fullscreen res: %ux%u", screen->w, screen->h);
				texttimer = 240; //unit of measure is frames (at 60 FPS)
			}
		}
		while (paused == true) {
			processinput();
			SDL_Delay(10);
			if (dosavestate == true) savestate();
			if (doloadstate == true) loadstate();
		}
		if (dosavestate == true) savestate();
		if (doloadstate == true) loadstate();
	}

	getclock(endtimer);

	/*difftimer = ((endtimer * 1000) - (starttimer * 1000)) / timerfreq;
	doubletimer = (double)difftimer / (double)1000.0;
	printf("\n");
    printf("Executed %lu CPU clock ticks in %f seconds. (approx. %lu clock ticks/sec)\n", getclockticks(), doubletimer, (uint32_t)((double)getclockticks() / doubletimer));
	printf("Rendered %lu frames.\n", totalframes);*/
}

void timedtext(uint8_t *text) {
	sprintf(&persistenttext[0], text);
	texttimer = 240; //unit of measure is frames (at 60 FPS)
}
