/* MoarNES source - main.c

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   MoarNES is an emulator of the Nintendo Entertainment System (NES) game
   console that was very popular in the 1980's. It emulates the various
   components of the console, such as the CPU, graphics and audio processors,
   memory mapping of the system, etc.

   It makes use of the Simple Directmedia Layer library by Sam Latinga, for
   ease is porting this emulator to various platforms without needing to
   completely rewrite any low-level graphics, audio, or input code.

   All other code in this emulator is written completely by Mike Chambers.

   It can run dumped images of most of the original NES game cartridges. These are
   widely available on the internet.
*/


#include <stdio.h>
#include "config.h"
#include "moarnes.h"
#ifndef _WIN32
#include <X11/Xlib.h>
#endif

const uint8_t *build = VERSION_STRING;
bool ROMwaiting = false;
bool exitapp = false;
uint8_t nextROM[512];
uint8_t app_path[MAX_PATH];
uint8_t doaudio = 1;

#ifdef _WIN32
	HANDLE GameThread;
	HANDLE NESThread;
	HANDLE MainLoopThread;
#endif

extern uint8_t dorecord;
extern uint8_t recordpath[256];
extern uint8_t outputNES[240][256];

extern bool running;
extern bool paused;
extern uint32_t cursamplerate;
extern uint16_t changescreenmode;

void mainloop(void *dummy) {
	while(exitapp == false) {
		if (running == true) {
			execNES();
		} else if (ROMwaiting == true) {
			loadROM(nextROM);
			running = true;
		} else if (changescreenmode) {
			setscreenmode(changescreenmode);
			changescreenmode = 0;
		}
		SDL_Delay(1);
	}
}

bool audioready = false;
void audiothread(void *dummy) {
	if (initAPU(1, 48000) == false) if (initAPU(1, 44100) == false)
		if (initAPU(1, 32000) == false) if (initAPU(1, 22050) == false) if (initAPU(1, 11025) == false) doaudio = 0;
	audioready = true;
	while(exitapp == false) {
		SDL_Delay(100);
	}
}

int main(int argc, char *argv[]) {
	uint8_t textdata[256] = { 0 };
	int16_t charfind;

#ifdef _WIN32
	charfind = (int16_t)GetModuleFileName(NULL, (LPSTR)app_path, MAX_PATH);
	for (; charfind>=0; charfind--) {
		if (app_path[charfind] == '\\') {
			app_path[charfind+1] = 0;
			break;
		}
	}
#else
	pthread_t mainthread;
	XInitThreads();
	sprintf(app_path, "./");
#endif

	sprintf((int8_t *)textdata, "MoarNES %s", build);
	if (initSDL(textdata) == false) return(-1);
	if (initAPU(1, 48000) == false) if (initAPU(1, 44100) == false)
		if (initAPU(1, 32000) == false) if (initAPU(1, 22050) == false) if (initAPU(1, 11025) == false) doaudio = 0;
	initPPU();

	hookexternal(&tickchannelsAPU);
	initinput();
	loadconfig();

	memset((void *)outputNES, 65, 61440); //blank screen to black
	drawtext(textdata, (uint16_t)(128-(strlen((int8_t *)textdata)*8)/2), 224);
	if (doaudio)
		sprintf((int8_t *)textdata, "Audio initialized at %lu Hz", cursamplerate);
	else
		sprintf((int8_t *)textdata, "Unable to initialize audio");
	drawtext(textdata, (uint16_t)(128-(strlen((int8_t *)textdata)*8)/2), 232);
	blitframe((uint8_t *)outputNES);

#ifdef _WIN32
	_beginthread(mainloop, 0, NULL);
#else
	pthread_create(&mainthread, NULL, (void *)mainloop, NULL);
	if (!loadROM(argv[1])) return(-1);
	ROMwaiting = true;
	running = true;
	if (argc > 2) setscreenmode(atoi(argv[2]));
#endif
	while(exitapp == false) {
		processinput();
		SDL_Delay(10);
	}

	saveconfig();
	return(0);
}

#ifdef _WIN32
extern HWND myWindow;
int16_t msgbox(uint8_t *msgtxt, uint16_t buttons) {
	return(MessageBox(myWindow, (LPCSTR)msgtxt, "MoarNES", buttons));
}

void fatalerr(uint8_t *errmsg) {
	uint8_t newmsg[1024];

	sprintf((int8_t *)newmsg, "FATAL ERROR: %s", errmsg);
	MessageBox(myWindow, (LPCSTR)newmsg, "Fatal error", MB_ICONERROR | MB_OK);
	exit(-1);
}
#else
int16_t msgbox(uint8_t *msgtxt, uint16_t buttons) {
//	return(MessageBox(myWindow, (LPCSTR)msgtxt, "MoarNES", buttons));
	if ((!buttons & MB_YESNO)) {
		printf(msgtxt);
		printf("\n");
	}
	return(IDYES);
}

void fatalerr(uint8_t *errmsg) {
	printf("FATAL ERROR: %s", errmsg);
	exit(-1);
}
#endif

bool verifyexit() {
	switch (msgbox("Are you sure you want to exit MoarNES?", MB_ICONQUESTION | MB_YESNO)) {
		case IDYES:
			running = false;
			paused = false;
			exitapp = true;
			return(true);
	}

	return(false);
}

extern uint8_t app_path[MAX_PATH];
extern struct inputdev_s gamepad[2];
extern uint16_t screenmode;
#ifdef _WIN32
extern HWND myWindow;
RECT winrect;
#else
uint8_t blank[4] = { 0, 0, 0, 0 };
#endif
void loadconfig() {
	FILE *configfile = NULL;
	uint8_t path[MAX_PATH];
	int16_t wx, wy;

	sprintf((int8_t *)path, "%sMoarNES.conf", app_path);
	configfile = fopen((int8_t *)path, "rb");
	if (configfile == NULL) {
		msgbox("This appears to be the first time you have run MoarNES, or the configuration file has been deleted. A new one will be created in the application's folder using the default options.", MB_ICONINFORMATION | MB_OK);
		saveconfig();
		return;
	}

	fread(&screenmode, 1, 2, configfile);
	fread(&wx, 1, 2, configfile);
	fread(&wy, 1, 2, configfile);
	#ifdef _WIN32
		SetWindowPos(myWindow, NULL, wx, wy, 0, 0, SWP_NOSIZE);
	#endif
	fread((void *)gamepad, 1, sizeof(struct inputdev_s) << 1, configfile);
	fclose(configfile);

	setscreenmode(screenmode);
}

void saveconfig() {
	FILE *configfile = NULL;
	uint8_t path[MAX_PATH];

	sprintf((int8_t *)path, "%sMoarNES.conf", app_path);
	configfile = fopen((int8_t *)path, "wb");
	if (configfile == NULL) return;

	fwrite(&screenmode, 1, 2, configfile);
	#ifdef _WIN32
		GetWindowRect(myWindow, (LPRECT)&winrect);
		fwrite(&winrect.left, 1, 2, configfile);
		fwrite(&winrect.top, 1, 2, configfile);
	#else
		fwrite((void *)blank, 1, 4, configfile);
	#endif
	fwrite((void *)gamepad, 1, sizeof(struct inputdev_s) << 1, configfile);

	fclose(configfile);
}
